class MidasClientError(Exception):
    """Exception to indicate a general API error."""


class MidasCommunicationError(MidasClientError):
    """Exception to indicate a communication error."""


class MidasAuthenticationError(MidasClientError):
    """Exception to indicate an authentication error."""



from dataclasses import dataclass, field
from typing import Optional
from datetime import datetime
from dateutil import parser

@dataclass
class RateListItem():
    RateID: str
    SignalType: str
    Description: str
    LastUpdated: Optional[str] = None

@dataclass
class ValueInfoItem:
    ValueName: str
    DateStart: str
    DateEnd: str
    DayStart: str
    DayEnd: str
    TimeStart: str
    TimeEnd: str
    value: float
    Unit: str

    def GetStart(self) -> datetime:
        """Get the start of this tariff as a python datetime"""
        return parser.parse(f"{self.DateStart} {self.TimeStart} -07:00")


    def GetEnd(self) -> datetime:
        """Get the end of this tariff as a python datetime"""
        return parser.parse(f"{self.DateEnd} {self.TimeEnd} -07:00")

@dataclass
class RateInfo:
    RateID: str
    SystemTime_UTC: str
    RateName: str
    RateType: str
    Sector: str
    API_Url: str
    RatePlan_Url: str
    EndUse: str
    AltRateName1: str
    AltRateName2: str
    SignupCloseDate: str
    ValueInformation: list[ValueInfoItem] = field(default_factory=list)
    """The list of tariffs"""

    def GetCurrentTariffs(self) -> list[ValueInfoItem]:
        """Gets all tariffs that are currently active"""
        
        # ValueInformation where getstart < now < getend
        now = datetime.now().timestamp()
        return list(filter(lambda t: t.GetStart().timestamp() < now and t.GetEnd().timestamp() > now, self.ValueInformation))


from .authentication import Midas as MidasAuth
from .ratelist import Midas as MidasRateList

class Midas(MidasAuth, MidasRateList):
    """
    Python API for California's energy price database MIDAS.

    Main class, imports most things from modules.
    """


from typing import Literal
import requests
import base64
import jwt
import time

from .exception import MidasAuthenticationError

class Midas():
    def __request(self, method: Literal['GET', 'POST'], url: str):
        """Preform a request with the stored auth token and return the body."""
        if (self.auth_token is None or not Midas.__isTokenValid(self.auth_token)):
            self.__loginAndStore(self.username, self.password)
        headers = {
            'Accept': 'application/json',
            'User-Agent': 'california-midasapi.py',
            'Authorization': "Bearer " + self.auth_token,
        }
        response = requests.request(method, url, headers=headers)
        if (not response.ok):
            raise f"Error preforming request: {response.status_code} {response.text}"
        return response.text
    
    def __loginAndStore(self, username: str, password: str):
        '''
        Logs in with a username and password, storing the JWT token for use in future calls.
        '''
        credentials = username + ":" + password
        credentials_encodedBytes = base64.b64encode(credentials.encode("utf-8"))
    
        headers = {b'Authorization': b'BASIC ' + credentials_encodedBytes}
        url = 'https://midasapi.energy.ca.gov/api/token'

        response = requests.get(url,headers=headers)

        if (not response.ok):
            raise MidasAuthenticationError(response.text)

        self.auth_token = response.headers['Token']
    
    @staticmethod
    def __isTokenValid(token: str) -> bool:
        """Return if the provided token is still valid"""
        decoded = jwt.decode(token, algorithms=["HS256"], options={"verify_signature": False})
        future = time.time() + 120 # 2 minutes from now
        return decoded["exp"] > future # expires more than set time from now


from .internal import Midas as Internal
from .types import RateListItem, ValueInfoItem, RateInfo
from typing import Literal
from enum import Enum
import json

class RINFilter(Enum):
        """Filter options for the RIN list."""
        ALL = 0
        TARIFF = 1
        GHG_EMISSION = 2
        FLEX_ALERT = 3

class Midas(Internal):

    def GetAvailableRates(self, signaltype: RINFilter) -> 'list[RateListItem]':
        """
        Get all the available rates.
        """

        def __rateListItemHook(dict):
             return RateListItem(**dict)

        url = 'https://midasapi.energy.ca.gov/api/valuedata?signaltype=' + str(signaltype.value)
        response = self.__request('GET', url)
        return (json.loads(response, object_hook=__rateListItemHook))
    
    def GetRateInfo(self, rateID: str, queryType: Literal['alldata', 'realtime'] = 'alldata') -> RateInfo:
        """
        Returns data about a given a rate.
        """

        def __rateInfoObjectHook(dict):
            # this has to handle both the parent object and its children
            if "DateStart" in dict:
                return ValueInfoItem(**dict)
            elif "RateID" in dict:
                return RateInfo(**dict)
            else:
                raise "Invalid object type for __rateInfoObjectHook"
            
        # TODO what does queryType=realtime even do? all properties are None
        url = 'https://midasapi.energy.ca.gov/api/valuedata?id=' + rateID + '&querytype=' + queryType
        pricing_response = self.__request('GET', url)

        return (json.loads(pricing_response, object_hook=__rateInfoObjectHook))


import requests
import json
import base64
from .internal import Midas as Internal

class Midas(Internal):
    def __init__(self, username: str, password: str):
        """
        Create a new API wrapper instance using the given credentials.

        Credentials are required, if you don't have an account use the static `register` method
        """
        self.username = username
        self.password = password
        self.auth_token: str = None
    
    def test_credentials(self) -> bool:
        """Test the provided credentials. Throws if invalid or expired."""
        self.__loginAndStore(self.username, self.password)
    
    @staticmethod
    def register(username: str, password: str, email: str, fullname: str, organization: str = None) -> str:
        """
        Create a new account with the MIDAS server.

        
        """
        username64 = str(base64.b64encode(username.encode("utf-8")), "utf-8")
        password64 = str(base64.b64encode(password.encode("utf-8")), "utf-8")
        email64 = str(base64.b64encode(email.encode("utf-8")), "utf-8")
        fullname64 = str(base64.b64encode(fullname.encode("utf-8")), "utf-8")

        registration_info = {
            "username":username64,
            "password":password64,
            "emailaddress":email64,
            "fullname":fullname64
        }

        if (organization is not None):
            organization64 = str(base64.b64encode(organization.encode("utf-8")), "utf-8")
            registration_info["organization"] = organization64

        url = 'https://midasapi.energy.ca.gov/api/registration'
        headers =  {"Content-Type":"application/json"}

        response = requests.post(url, data=json.dumps(registration_info), headers=headers)

        #Prints below will return 200 response for successful call
        print(response)
        #Response text should be: 'User account for your_user_name was successfully created. A verification email has been sent to your_email. Please click the link in the email in order to start using the API.'
        print(response.text)
        return response.text

